/* 
 * Copyright (C) 2020 Raven Computing
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.raven.research.bitspark;

import java.io.IOException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;
import java.util.TreeSet;

import com.raven.common.io.CSVFileWriter;
import com.raven.common.io.DataFrameSerializer;
import com.raven.common.struct.DataFrame;
import com.raven.common.util.ArgumentParseException;
import com.raven.common.util.ArgumentParser;
import com.raven.common.util.Chronometer;
import com.raven.research.bitspark.subj.RavenTestSubject;
import com.raven.research.bitspark.subj.Subject;
import com.raven.research.bitspark.util.Log;
import com.raven.research.bitspark.util.TestLap;
import com.raven.research.bitspark.util.TestRunner;

/**
 * Main class for creating Bloom filter benchmarks.
 *
 */
public class Benchmarker {
    
    public static void main(String[] args){
        final ArgumentParser ap = parseArgs(args);
        Log.info("---------- Benchmarker ----------");
        
        final String selectedSubject = ap.getStringArg("subject");
        final String fileFormat = ap.getStringArg("fileFormat", "df");
        //Set initial capacity flag.
        //This is only used by our scalable implementation
        RavenTestSubject.useMinimizedInitCapacity = ap.getBooleanArg("minimizedInitCap", false);
        
        //Either create test laps for a specific subject or for all subjects
        final List<TestLap> laps = ((selectedSubject != null) && (!selectedSubject.isEmpty()))
                        ? TestLap.createFor(Subject.byName(selectedSubject))
                        : TestLap.createFor(Subject.values());
                        
        final Chronometer chrono = new Chronometer();
        chrono.start();
        final DataFrame result = TestRunner.run(laps);
        chrono.stop();
        save(result, fileFormat);
        Log.summary(laps.size(), chrono);
        show(result); 
    }
    
    private static void save(final DataFrame result, final String fileFormat){
        final String time = DateTimeFormatter.ofPattern("yyyyMMdd-HHmmss")
                .withZone(ZoneId.systemDefault()).format(Instant.now());
                
        final String file = "benchmark-" + time;
        try{
            if(fileFormat.equalsIgnoreCase("df")){
                DataFrameSerializer.writeFile(file, result);
            }else{
                new CSVFileWriter(file + ".csv").write(result);
            }
        }catch(IOException ex){
            Log.error("An error occurred while trying to persist the results to a file");
            Log.error(ex);
        }
    }
    
    private static void show(final DataFrame result){
        Log.info();
        Log.info("Would you like to view the results? (Y/n) ");
        System.out.print("       ");
        final Scanner sc = new Scanner(System.in);
        final String line = sc.nextLine();
        sc.close();
        if((line != null) && (!line.isEmpty())){
            if(!line.equalsIgnoreCase("N") && !line.equalsIgnoreCase("No")){
                Log.info();
                Log.info();
                System.out.println(result);
            }
        }
        System.out.println();
    }
    
    private static ArgumentParser parseArgs(final String[] args){
        final ArgumentParser ap = new ArgumentParser.Builder()
                .stringArg("subject",  new TreeSet<>(Subject.getAllNames()), true)
                .stringArg("fileFormat",  new TreeSet<>(Arrays.asList("df", "csv")), true)
                .booleanArg("debug", true)
                .booleanArg("useAnsiColors", true)
                .booleanArg("minimizedInitCap", true)
                .build();
        
        try{
            ap.parse(args);
        }catch(ArgumentParseException ex){
            Log.error("Invalid argument: " + ex.cause());
            Log.info();
            Log.info(ex.hint());
            Log.info();
            System.exit(1);
        }
        
        Log.debug = ap.getBooleanArg("debug", false);
        Log.useAnsiColors = ap.getBooleanArg("useAnsiColors", true);
        return ap;
    }
}
