/* 
 * Copyright (C) 2020 Raven Computing
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.raven.research.bitspark.struct;

import java.math.RoundingMode;
import java.nio.charset.StandardCharsets;
import java.text.DecimalFormat;

import com.google.common.math.DoubleMath;

/**
 * An implementation of a scalable Bloom filter.
 * 
 * @author Phil Gaiser
 *
 */
public class BloomFilter {
    
    /**
     * The default scale factor of a Bloom filter
     */
    public static final int DEFAULT_SCALE_FACTOR = 4;
    
    /**
     * The default tightening ration of a Bloom filter
     */
    public static final double DEFAULT_TIGHTENING_RATION = 0.9;
    
    private Stage[] stages;
    private final int scaleFactor;
    private final double tighteningRatio;
    private final double fpr0;
    
    /**
     * Constructs a new <code>BloomFilter</code> with the specified initial
     * capacity and maximum allowed false positive probability.<br>
     * The Bloom filter will have a default scale factor and tightening ration
     * 
     * @param initialCapacity The number of elements the
     *                        constructed BloomFilter should be able to hold
     * @param maxError The maximum allowed false positive probability the
     *                  constructed BloomFilter should adhere to
     */
    public BloomFilter(final int initialCapacity, final double maxError){
        this(initialCapacity, maxError, DEFAULT_SCALE_FACTOR, DEFAULT_TIGHTENING_RATION);
    }
    
    /**
     * Constructs a new <code>BloomFilter</code> with the specified initial
     * capacity, maximum allowed false positive probability and scale factor.<br>
     * The Bloom filter will have a default tightening ration
     * 
     * @param initialCapacity The number of elements the
     *                        constructed BloomFilter should be able to hold
     * @param maxError The maximum allowed false positive probability the
     *                  constructed BloomFilter should adhere to
     * @param scaleFactor The scale factor of the constructed BloomFilter. Usually either 2 or 4
     */
    public BloomFilter(final int initialCapacity, final double maxError, final int scaleFactor){
        this(initialCapacity, maxError, scaleFactor, DEFAULT_TIGHTENING_RATION);
    }
    
    /**
     * 
     * 
     * @param initialCapacity The number of elements the
     *                        constructed BloomFilter should be able to hold
     * @param maxError The maximum allowed false positive probability the
     *                  constructed BloomFilter should adhere to
     * @param scaleFactor The scale factor of the constructed BloomFilter.
     *                    Usually either 2 or 4
     * @param tighteningRatio The tightening ration of the constructed BloomFilter.
     *                        Usually between 0.8 and 0.9
     */
    public BloomFilter(final int initialCapacity, final double maxError,
            final int scaleFactor, final double tighteningRatio){
        
        this.fpr0 = maxError;
        this.scaleFactor = scaleFactor;
        this.tighteningRatio = tighteningRatio;
        this.stages = new Stage[1];
        this.stages[0] = new Stage(initialCapacity, maxError);
    }
    
    /**
     * Adds the specified element to this BloomFilter
     * 
     * @param str The String to be added to this BloomFilter
     */
    public void add(final String str){
        final long hash = hash(str);
        final Stage stage = stages[stages.length - 1];
        if(stage.isFull()){
            final Stage newStage = new Stage(this,
                    (stage.capacity * scaleFactor),
                    (stage.fpr * tighteningRatio));
            
            ensureCapacity();
            stages[stages.length - 1] = newStage;
            newStage.put(hash);
        }else{
            stage.put(hash);
        }
    }

    /**
     * Indicates whether the specified element might have
     * been added to this BloomFilter. The returned value is subject
     * to a probability of false positive error. This means that a true
     * return value might be wrong and the specified element is not
     * actually in this BloomFilter. If the return value is false the the
     * specified element is guaranteed to not be in this BloomFilter
     * 
     * @param str The String to check for membership in this BloomFilter
     * @return True if the specified element might have been added to
     *         this BloomFilter, fasle if it is definitely not
     *         in this BloomFilter
     */
    public boolean contains(final String str){
        final long hash = hash(str);
        for(int i=stages.length-1; i>=0; --i){
            if(stages[i].contains(hash)){
                return true;
            }
        }
        return false;
    }
    
    @Override
    public String toString(){
        final StringBuilder sb = new StringBuilder();
        final String nl = System.lineSeparator();
        final DecimalFormat d = new DecimalFormat("0.000");
        sb.append("Mode=");
        sb.append(scaleFactor);
        sb.append(nl);
        sb.append("TighteningRatio=");
        sb.append(tighteningRatio);
        sb.append(nl);
        sb.append("Max FPR=");
        sb.append(fpr0);
        sb.append(nl);
        sb.append("Stages fill ratios (in bits):");
        sb.append(nl);
        long totalSize = 0L;
        for(int i=0; i<stages.length; ++i){
            final Stage s = stages[i];
            totalSize += s.filter.size();
            sb.append("S");
            sb.append(i);
            sb.append(": ");
            sb.append(s.filter.bitsSet());
            sb.append("/");
            sb.append(s.filter.size());
            sb.append(" (");
            sb.append(d.format(((float)s.filter.bitsSet() / (float)s.filter.size()) * 100));
            sb.append("% full)");
            sb.append(nl);
        }
        final long sizeBytes = (totalSize / 8);
        sb.append("Total size: ");
        sb.append(sizeBytes);
        sb.append(" bytes (");
        sb.append(sizeBytes / 1024);
        sb.append(" KB)");
        return sb.toString();
    }
    
    /**
     * Resizes the internal array to hold at least one more Stage
     */
    private void ensureCapacity(){
        final int length = stages.length;
        final Stage[] tmp = new Stage[length + 1];
        for(int i=0; i<length; ++i){
            tmp[i] = stages[i];
        }
        this.stages = tmp;
    }
    
    /**
     * Hashes the specified String value
     * 
     * @param str The String object to hash
     * @return A hash for the specified String value
     */
    private long hash(final String str){
        return hashMurmur3(str.getBytes(StandardCharsets.UTF_8));
    }
    
    /**
     * Computes the Murmur3 hash value from the specified bytes.<br>
     * The original implementation of this method
     * was taken from Google's Guava library which is a
     * port of Austin Applebys original C++ code
     */
    private static long hashMurmur3(final byte[] data){
        long h1 = 0;
        long h2 = 0;
        int ptr = -1;
        final int blocks = (data.length / 16);
        for(int i = 0; i<blocks; ++i){
            long k1 = (((long) (data[++ptr]) << 56)
                     | ((long) (data[++ptr] & 0xff) << 48)
                     | ((long) (data[++ptr] & 0xff) << 40)
                     | ((long) (data[++ptr] & 0xff) << 32)
                     | ((long) (data[++ptr] & 0xff) << 24)
                     | ((long) (data[++ptr] & 0xff) << 16)
                     | ((long) (data[++ptr] & 0xff) << 8)
                     | ((long) (data[++ptr] & 0xff)));

            long k2 = (((long) (data[++ptr]) << 56)
                     | ((long) (data[++ptr] & 0xff) << 48)
                     | ((long) (data[++ptr] & 0xff) << 40)
                     | ((long) (data[++ptr] & 0xff) << 32)
                     | ((long) (data[++ptr] & 0xff) << 24)
                     | ((long) (data[++ptr] & 0xff) << 16)
                     | ((long) (data[++ptr] & 0xff) << 8)
                     | ((long) (data[++ptr] & 0xff)));

            h1 ^= mixK1(k1);
            
            h1 = Long.rotateLeft(h1, 27);
            h1 += h2;
            h1 = h1 * 5 + 0x52dce729;
            
            h2 ^= mixK2(k2);

            h2 = Long.rotateLeft(h2, 31);
            h2 += h1;
            h2 = h2 * 5 + 0x38495ab5;
        }
        final int remaining = (data.length % 16);
        if(remaining > 0){
            ++ptr;
            long k1 = 0;
            long k2 = 0;
            switch (remaining){
                case 15:
                    k2 ^= (long) (data[ptr + 14] & 0xff) << 48;
                    
                case 14:
                    k2 ^= (long) (data[ptr + 13] & 0xff) << 40;
                
                case 13:
                    k2 ^= (long) (data[ptr + 12] & 0xff) << 32;
                
                case 12:
                    k2 ^= (long) (data[ptr + 11] & 0xff) << 24;
                
                case 11:
                    k2 ^= (long) (data[ptr + 10] & 0xff) << 16;
                
                case 10:
                    k2 ^= (long) (data[ptr + 9] & 0xff) << 8;
                
                case 9:
                    k2 ^= (long) (data[ptr + 8] & 0xff);
        
            case 8:
                k1 ^= (((long) (data[ptr++]) << 56)
                     | ((long) (data[ptr++] & 0xff) << 48)
                     | ((long) (data[ptr++] & 0xff) << 40)
                     | ((long) (data[ptr++] & 0xff) << 32)
                     | ((long) (data[ptr++] & 0xff) << 24)
                     | ((long) (data[ptr++] & 0xff) << 16)
                     | ((long) (data[ptr++] & 0xff) << 8)
                     | ((long) (data[ptr++] & 0xff)));

                break;
                case 7:
                    k1 ^= (long) (data[ptr + 6] & 0xff) << 48;
                
                case 6:
                    k1 ^= (long) (data[ptr + 5] & 0xff) << 40;
                
                case 5:
                    k1 ^= (long) (data[ptr + 4] & 0xff) << 32;
                
                case 4:
                    k1 ^= (long) (data[ptr + 3] & 0xff) << 24;
                
                case 3:
                    k1 ^= (long) (data[ptr + 2] & 0xff) << 16;
                
                case 2:
                    k1 ^= (long) (data[ptr + 1] & 0xff) << 8;
                
                case 1:
                    k1 ^= (long) (data[ptr] & 0xff);
                    break;
                    
                default:
                    throw new AssertionError("Implementation Error");
            }
            h1 ^= mixK1(k1);
            h2 ^= mixK2(k2);
        }
        h1 += h2;
        h2 += h1;

        h1 = fmix64(h1);
        h2 = fmix64(h2);

        h1 += h2;
        h2 += h1;
        
        return h1;
    }
    
    private static long mixK1(long k1) {
        k1 *= 0x87c37b91114253d5L;
        k1 = Long.rotateLeft(k1, 31);
        k1 *= 0x4cf5ad432745937fL;
        return k1;
    }

    private static long mixK2(long k2) {
        k2 *= 0x4cf5ad432745937fL;
        k2 = Long.rotateLeft(k2,  33);
        k2 *= 0x87c37b91114253d5L;
        return k2;
    }
    
    private static long fmix64(long k) {
        k ^= k >>> 33;
        k *= 0xff51afd7ed558ccdL;
        k ^= k >>> 33;
        k *= 0xc4ceb9fe1a85ec53L;
        k ^= k >>> 33;
        return k;
    }
    
    /**
     * Models a size immutable stage of a Bloom filter allowing
     * it to grow dynamically as a whole.
     *
     */
    private static class Stage {
        
        private final BitVector filter;
        private final double fpr;
        private final int slices;
        private final int sliceSize;
        private final int capacity;
        
        Stage(final int capacity, final double errorRate){
            this.capacity = capacity;
            this.fpr = errorRate;
            this.slices = DoubleMath.log2((1 / errorRate), RoundingMode.CEILING);
            this.sliceSize = (int) Math.ceil((capacity * Math.abs(
                    Math.log(errorRate))) / (slices * 0.480453014));//const: ln(2)^2
            
            final int vectorSize = (slices * sliceSize);
            this.filter = BitVector.createInitialized(vectorSize, false);
        }
        
        Stage(final BloomFilter parent, final int capacity, final double errorRate){
            this.capacity = capacity;
            this.fpr = errorRate;
            final int mode = parent.scaleFactor;
            final int m0 = parent.stages[0].sliceSize;
            final int k0 = parent.stages[0].slices;
            final int i = parent.stages.length + 1;
            final double tighteningRatio = 0.9;
            this.slices = (int) Math.round(Math.ceil(k0 + (i *
                    DoubleMath.log2((1 / tighteningRatio), RoundingMode.CEILING))));
            
            this.sliceSize = (int) Math.ceil((m0 * (Math.pow(mode, i - 1))));
            final int vectorSize = (slices * sliceSize);
            this.filter = BitVector.createInitialized(vectorSize, false);
        }
        
        /**
         * Adds the specified hash to this Stage
         * 
         * @param hash The hash value to be added to this Stage
         */
        void put(final long hash){
            int h1 = (int) hash;
            final int h2 = (int) (hash >>> 32);
            int offset = 0;
            for(int i=0; i<slices; ++i){
                filter.set(((h1 & 0x7fffffff) % sliceSize) + offset, true);
                h1 += h2;
                offset += sliceSize;
            }
        }
        
        /**
         * Indicates whether the specified hash might have
         * been added to this Stage
         * 
         * @param hash The hash value to check
         * @return True if the specified hash value might have been added to
         *         this Stage, false if it is definitely not in this Stage
         */
        boolean contains(final long hash){
            int h1 = (int) hash;
            final int h2 = (int) (hash >>> 32);
            int offset = 0;
            for(int i=0; i<slices; ++i){
                if(!filter.get(((h1 & 0x7fffffff) % sliceSize) + offset)){
                    return false;
                }
                h1 += h2;
                offset += sliceSize;
            }
            return true;
        }
        
        /**
         * Indicates whether this Stage has reached its maximum allowed capacity,
         * i.e. its fill ration
         * 
         * @return True if this Stage's fill ration is over the maximum allowed value,
         *         false otherwise
         */
        boolean isFull(){
            return (filter.bitsSet() > (filter.size() / 2));
        }
        
        @Override
        public String toString(){
            final StringBuilder sb = new StringBuilder();
            final String nl = System.lineSeparator();
            sb.append(filter.toString());
            sb.append(nl);
            return sb.toString();
        }
    }
    
}
