/* 
 * Copyright (C) 2020 Raven Computing
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.raven.research.bitspark.subj;

import java.util.Arrays;
import java.util.List;

/**
 * This enum lists all test subjects for benchmarking
 * various Bloom filter implementations.
 *
 */
public enum Subject {
    
    /**
     * Our implementation of a scalable Bloom filter.
     */
    RAVEN_COMPUTING(RavenTestSubject.class, "Raven Computing"),
    
    /**
     * An implementation of a static Bloom filter provided by Baqend.<br>
     * See their
     * <a href="https://github.com/Baqend/Orestes-Bloomfilter">repository</a>
     * for more information.
     */
    BAQEND(BaqendTestSubject.class, "Baqend"),
    
    /**
     * An implementation of a static Bloom filter provided by Google.<br>
     * See the
     * <a href="https://github.com/google/guava">repository</a> of the Guava
     * library for more information.
     */
    GOOGLE(GoogleTestSubject.class, "Google"),
    
    /**
     * An implementation of a static Bloom filter provided by Sandeep Gupta (Sangupta).<br>
     * See his
     * <a href="https://github.com/sangupta/bloomfilter">repository</a>
     * for more information.
     */
    SANGUPTA(SanguptaTestSubject.class, "Sangupta"),
    
    /**
     * An implementation of a Cuckoo filter provided by Mark Gunlogson (Gunlogson).<br>
     * See his
     * <a href="https://github.com/MGunlogson/CuckooFilter4J">repository</a>
     * for more information.
     */
    GUNLOGSON(GunlogsonTestSubject.class, "Gunlogson");
    
    private Class<? extends TestSubject> implClass;
    private String devName;
    
    Subject(final Class<? extends TestSubject> implClass,
            final String devName){
        
        this.implClass = implClass;
        this.devName = devName;
    }
    
    public Class<? extends TestSubject> getImplementingClass(){
        return this.implClass;
    }
    
    public String getName(){
        return this.devName;
    }
    
    @Override
    public String toString(){
        return this.getName();
    }
    
    public static List<Subject> getAllSubjects(){
        return Arrays.asList(Subject.values());
    }
    
    public static List<String> getAllNames(){
        final Subject[] subjects = values();
        final String[] names = new String[subjects.length];
        for(int i=0; i<subjects.length; ++i){
            names[i] = subjects[i].getName();
        }
        return Arrays.asList(names);
    }
    
    public static Subject byName(final String devName){
        final Subject[] subjects = values();
        for(final Subject sbj : subjects){
            if(sbj.getName().equalsIgnoreCase(devName)){
                return sbj;
            }
        }
        throw new IllegalStateException("No Subject found for developer name: " + devName);
    }
    
}
