/* 
 * Copyright (C) 2020 Raven Computing
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.raven.research.bitspark.util;

import java.security.SecureRandom;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

/**
 * Models a dataset used for benchmarks and provides
 * a utility method for creating new datasets.
 *
 */
public final class Dataset {
    
    /**
     * Enumerates the different types of datasets.
     *
     */
    public enum Type {
        
        PATTERN,
        RANDOM;
        
        @Override
        public String toString(){
            final String s = super.toString();
            return Character.toUpperCase(s.charAt(0))
                    + s.toLowerCase().substring(1, s.length());
        }

    }
    
    static {
        final SecureRandom sr = new SecureRandom();
        rand1 = new Random(sr.nextLong());
        rand2 = new Random(sr.nextLong());
        rand3 = new Random(sr.nextLong());
    }
    
    public static final int SIZE_NEGATIVES = 200000;
    
    public static final int RANDOM_WORD_LENGTH = 16;
    
    public static final char[] RANDOM_ALPHABET = {
            '1','2','3','4','5','6','7','8','9','0',
            'a','b','c','d','e','f','g','h','i','j',
            'k','l','m','n','o','p','q','r','s','t',
            'u','v','w','x','y','z','A','B','C','D',
            'E','F','G','H','I','J','K','L','M','N',
            'O','P','Q','R','S','T','U','V','W','X',
            'Y','Z'};
    
    private static final Random rand1;
    private static final Random rand2;
    private static final Random rand3;
    
    private Set<String> data;
    private Set<String> negatives;
    
    private Dataset(final Set<String> data, final Set<String> negatives){
        this.data = data;
        this.negatives = negatives;
    }
    
    public Set<String> getData(){
        return this.data;
    }

    public Set<String> getNegatives(){
        return this.negatives;
    }

    public static Dataset create(final int size, final Dataset.Type type){
        if(type == null){
            throw new IllegalArgumentException("Dataset type argument must not be null");
        }
        switch(type){
        case PATTERN:
            return createPatternData(size);
        case RANDOM:
            return createRandomData(size);
        default:
            throw new IllegalStateException("Unknown enum type: " + type);
        }
    }

    private static Dataset createRandomData(final int size){
        Log.debug("Creating RANDOM dataset with size " + size);
        final Set<String> set = createSet();
        for(int i=0; i<size; ++i){
            set.add(createRandomWord());
        }
        Log.debug("Creating " + SIZE_NEGATIVES + " negative samples for RANDOM dataset");
        final Set<String> negatives = createSet();
        for(int i=0; i<SIZE_NEGATIVES; ++i){
            String word = createRandomWord();
            while(set.contains(word)){
                word = word + RANDOM_ALPHABET[rand1.nextInt() % RANDOM_ALPHABET.length];
            }
            negatives.add(createRandomWord());
        }
        return new Dataset(set, negatives);
    }

    private static Dataset createPatternData(final int size){
        Log.debug("Creating PATTERN dataset with size " + size);
        final Set<String> set = createSet();
        final int[] nums = new int[4];
        for(int i=0; i<size; ++i){
            int j = 0;
            while(nums[j] == 255){
                nums[j++] = 0;
            }
            nums[j] += 1;
            set.add(nums[0] + "." + nums[1] + "." + nums[2] + "." + nums[3]);
        }
        Log.debug("Creating " + SIZE_NEGATIVES + " negative samples for PATTERN dataset");
        final Set<String> negatives = createSet();
        final int length = ((size + 1) + SIZE_NEGATIVES);
        for(int i=(size + 1); i<length; ++i){
            int j = 0;
            while(nums[j] == 255){
                nums[j++] = 0;
            }
            nums[j] += 1;
            negatives.add(nums[0] + "." + nums[1] + "." + nums[2] + "." + nums[3]);
        }
        return new Dataset(set, negatives);
    }
    
    private static String createRandomWord(){
        final int alphabetLength = RANDOM_ALPHABET.length;
        final StringBuilder sb = new StringBuilder();
        for(int j=0; j<RANDOM_WORD_LENGTH; ++j){
            final int index = ((((rand1.nextInt() + rand2.nextInt()) ^ rand3.nextInt())
                                        & Integer.MAX_VALUE) % alphabetLength);
            
            sb.append(RANDOM_ALPHABET[index]);
        }
        return sb.toString();
    }
    
    private static Set<String> createSet(){
        return new TreeSet<String>();
    }
    
}
