/* 
 * Copyright (C) 2020 Raven Computing
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.raven.research.bitspark.util;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.raven.research.bitspark.subj.Subject;
import com.raven.research.bitspark.subj.TestSubject;
import com.raven.research.bitspark.subj.TestSubjectProvider;
import com.raven.research.bitspark.util.Dataset.Type;

/**
 * Represents a test lap for performing benchmarks.
 *
 */
public class TestLap {
    
    /**
     * The maximum false positives to be tested
     */
    public static final double[] TESTED_FPPS = {0.01, 0.001, 0.0001};
    
    /**
     * The dataset sizes to be tested
     */
    public static final int[] TESTED_SET_SIZES = {
                                     1000, 2000, 4000, 8000, 16000,
                                     32000, 64000, 128000, 256000, 512000,
                                     1024000, 2048000, 4096000, 8192000, 16384000};
    
    private TestSubject subject;
    private Dataset dataset;
    private Type setType;
    private int setSize;
    private double allowedFPP;
    
    public TestLap(final TestSubject subject, final int setSize,
            final double allowedFPP, final Type setType){
        
        this.subject = subject;
        this.setSize = setSize;
        this.allowedFPP = allowedFPP;
        this.setType = setType;
    }
    
    public void start(){
        Log.debug("Starting test lap: subject=\"" + subject.getDeveloperName()
                 + "\" type=" + setType + " size=" + setSize + " FPP=" + allowedFPP);
        
        this.dataset = Dataset.create(setSize, setType);
        if(dataset.getData().size() != setSize){
            throw new IllegalStateException("Set size is not equal to specified size: expected="
                                           + setSize + " actual=" + dataset.getData().size());
        }
        if(dataset.getNegatives().size() != Dataset.SIZE_NEGATIVES){
            throw new IllegalStateException(
                    "Set size for negative samples is not equal to specified size: expected="
                    + Dataset.SIZE_NEGATIVES + " actual=" + dataset.getNegatives().size());
        }
    }
    
    public void done(){
        //Clear the reference to the dataset so
        //that it can be garbage collected
        this.dataset = null;
        System.gc();
    }
    
    public TestSubject getSubject(){
        return this.subject;
    }

    public Set<String> getDataset(){
        if(dataset == null){
            throw new IllegalStateException("You must call start() first");
        }
        return this.dataset.getData();
    }
    
    public Set<String> getDatasetNegatives(){
        if(dataset == null){
            throw new IllegalStateException("You must call start() first");
        }
        return this.dataset.getNegatives();
    }
    
    public Type getDatasetType(){
        return this.setType;
    }
    
    public int getDatasetSize(){
        return this.setSize;
    }
    
    public double getAllowedFPP(){
        return this.allowedFPP;
    }
    
    public static List<TestLap> createFor(final Subject... subjects){
        final List<TestLap> laps = new LinkedList<>();
        for(final Subject sbj : subjects){
            createFullRoundFor(sbj, laps);
        }
        return laps;
    }
    
    private static void createFullRoundFor(final Subject sbj, final List<TestLap> laps){
        for(final Dataset.Type type : Dataset.Type.values()){
            for(final double fpp : TESTED_FPPS){
                for(final int size : TESTED_SET_SIZES){
                    laps.add(new TestLap(TestSubjectProvider.create(sbj), size, fpp, type));
                }
            }
        }
    }
    
}
