/* 
 * Copyright (C) 2020 Raven Computing
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.raven.research.bitspark.util;

import java.text.DecimalFormat;
import java.util.List;

import com.raven.common.struct.DataFrame;
import com.raven.common.struct.DefaultDataFrame;
import com.raven.common.util.Chronometer;
import com.raven.research.bitspark.subj.TestSubject;

/**
 * Utility class for running test laps.
 *
 */
public final class TestRunner {
    
    /**
     * The number of times to perform runtime
     * measurements of the add() and contains() methods.
     * This is used to compute the minimum, maximum and average runtime
     */
    public static final int RUNTIME_DATA_CYCLE = 5;
    
    public static DataFrame run(final List<TestLap> laps){
        Log.info("Running a total of " + laps.size() + " laps");
        final DataFrame df = new DefaultDataFrame(TestResult.class);
        int i = 1;
        for(final TestLap lap : laps){
            Log.progress(i, laps.size());
            final TestResult result = TestRunner.run(lap);
            df.addRow(result);
            ++i;
        }
        return df;
    }
    
    public static TestResult run(final TestLap lap){
        lap.start();
        final TestSubject subject = lap.getSubject();
        final TestResult result = new TestResult(subject.getDeveloperName());
        result.setDatasetSize(lap.getDatasetSize());
        result.setDatasetType(lap.getDatasetType().toString());
        result.setAllowedErrorRate(lap.getAllowedFPP());
        
        final RuntimeData timeAdd = runAdd(lap);
        final RuntimeData timeContains = runContains(lap);
        final int errors = measureFalsePositives(lap);
        final double errorRate = (double)errors / (double)lap.getDatasetNegatives().size();
        Log.info("Total false positives " + errors + " ("
                + new DecimalFormat("0.000").format(errorRate*100) + "%) ");
        
        result.setRuntimeAddMin(timeAdd.getMinimun());
        result.setRuntimeAddAvg(timeAdd.getAverage());
        result.setRuntimeAddMax(timeAdd.getMaximum());
        
        result.setRuntimeContainsMin(timeContains.getMinimun());
        result.setRuntimeContainsAvg(timeContains.getAverage());
        result.setRuntimeContainsMax(timeContains.getMaximum());
        
        result.setErrorCount(errors);
        result.setErrorRate(errorRate);
        
        lap.done();
        return result;
    }
    
    private static RuntimeData runAdd(final TestLap lap){
        long min = RuntimeData.defaultMinimum();
        long max = RuntimeData.defaultMaximum();
        long sum = 0L;
        final TestSubject subject = lap.getSubject();
        for(int i=0; i<RUNTIME_DATA_CYCLE; ++i){
            subject.initializeFor(lap);
            final Chronometer chrono = new Chronometer();
            chrono.start();
            for(final String elem : lap.getDataset()){
                subject.add(elem);
            }
            final long t = chrono.stop().elapsedMillis();
            Log.info("Insertion " + (i+1) + "/" + RUNTIME_DATA_CYCLE
                    + " took " + chrono.toString());
            
            min = RuntimeData.selectMin(t, min);
            max = RuntimeData.selectMax(t, max);
            sum += t;
        }
        final long avg = RuntimeData.averageFrom(sum, RUNTIME_DATA_CYCLE);
        return new RuntimeData(min, avg, max);
    }
    
    private static RuntimeData runContains(final TestLap lap){
        long min = RuntimeData.defaultMinimum();
        long max = RuntimeData.defaultMaximum();
        long sum = 0L;
        final TestSubject subject = lap.getSubject();
        for(int i=0; i<RUNTIME_DATA_CYCLE; ++i){
            final Chronometer chrono = new Chronometer();
            chrono.start();
            for(final String elem : lap.getDataset()){
                subject.contains(elem);
            }
            final long t = chrono.stop().elapsedMillis();
            Log.info("Membership query " + (i+1) + "/" + RUNTIME_DATA_CYCLE
                    + " took " + chrono.toString());
            
            min = RuntimeData.selectMin(t, min);
            max = RuntimeData.selectMax(t, max);
            sum += t;
        }
        final long avg = RuntimeData.averageFrom(sum, RUNTIME_DATA_CYCLE);
        return new RuntimeData(min, avg, max);
    }
    
    private static int measureFalsePositives(final TestLap lap){
        final TestSubject subject = lap.getSubject();
        int errors = 0;
        for(final String elem : lap.getDatasetNegatives()){
            if(subject.contains(elem)){
                ++errors;
            }
        }
        return errors;
    }
}
